import sigrokdecode as srd

class ChannelError(Exception):
    pass

class Decoder(srd.Decoder):
    api_version = 3
    id = 'grm10'
    name = 'GRM10'
    longname = 'Wild/Leica Tachymeter Karteninterface'
    desc = 'seriell mit Takt, halbduplex'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['grm10']
    tags = ['Embedded/industrial']
    channels = (
        {'id': 'sel', 'name': 'SEL', 'desc': 'Card select'},
        {'id': 'clk', 'name': 'CLK', 'desc': 'Clock'},
        {'id': 'dat', 'name': 'DAT', 'desc': 'Bidi data'},
    )
    options = (
        {'id': 'inverted', 'desc': 'Invertiert?', 'default': 'nein',
            'values': ('nein', 'ja (via V40098)')},
    )
    annotations = (
        ('bits', 'DAT bits'),   # normale Bits
        ('extra', 'DAT extra'), # Sonderbits
        ('nibbles', 'DAT extra'),
        ('query', 'Anfrage'),
        ('answer', 'Antwort'),
        ('record','Datensatz'),
        ('aux','irgendwas'),
    )
    annotation_rows = (
        ('bits', 'Bits', (0,1,)),
        ('nibbles', 'Halbbytes', (2,)),
        ('data', 'Daten', (3,4,)),
        ('record','Datensatz',(5,6)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
# Bit-Daten
        self.ss_bit = -1    # Startzeit (ausgegenen wird's am Ende)
# Nibble-Daten
        self.bitcount = 0
        self.ss_nibble = -1    # Startzeit (ausgegenen wird's am Ende)
# 16-Bit-Adresse
        self.a = 0
# Byte-Daten
        self.d = 0
        self.ss_ad = -1
# Datensatz (Record)
        self.r = []
        self.ss_r = -1  # Start-Sample

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)    # das wirklich wichtige!
        self.inverted = self.options['inverted'] != 'nein'

    def bit_complete(self):
        if self.ss_bit >= 0: # Bit zu Ende?
            self.put(self.ss_bit,self.samplenum,self.out_ann,self.bit)
        self.ss_bit = -1

    def nibble_complete(self):
        if self.ss_nibble >= 0:
# assert self.bitcount % 4
            self.put(self.ss_nibble,self.samplenum,self.out_ann,[2,['%X' % self.nibble]])
        self.ss_nibble = -1
    
    def nibbleExpand(self, n):
        return '0123456789!+?-._'[n]    # Unterstrich statt Leerzeichen zur Verdeutlichung

    def ad_complete(self, ann):
        if self.ss_ad >= 0:
            self.put(self.ss_ad,self.samplenum,self.out_ann,ann)
        self.ss_ad = -1

    def r_wantlen(self):
        return 16 if len(self.r)>=4 and self.r[3]&0x80 or self.a >= 0x8000 else 8   # Doppelrecord bei gesetztem Bit 7

    def r_complete(self):
        if (self.ss_r >= 0 and len(self.r)==self.r_wantlen()):
          ann = ''
          addr = "%04X" % (self.a + 1 - len(self.r))
          if self.a >= 0x8000:
            for b in self.r:
              ann += "%02X " % b
            self.put(self.ss_r,self.samplenum,self.out_ann,[6,[addr+": "+ann,ann]])
          else:
            for i,b in enumerate(self.r):
                if i == 3:  # Der Index 3 spielt im Bytearray eine Sonderrolle
                    ann += '+' if b&0x20 else '-'
                elif i >= 4 and self.r[3]&0x80:
                    if i<8 or i>= 12:   # zweiten Record übergehen, die ersten 4 Bytes sollten identisch sein!
                        ann += chr(b)   # 8 ASCII-Zeichen Kommentar erfordern Doppelrecord
                else:
                    ann += self.nibbleExpand(b>>4) + self.nibbleExpand(b&15)
            ann += '.' if self.r[3]&0x10 else '_'
            if self.r[3]&0x40:
                ann += '\\n'
            self.put(self.ss_r,self.samplenum,self.out_ann,[5,[addr+": „"+ann+"“","„"+ann+"“",ann]])
          self.ss_r = -1

    def clk_rise(self, dat):
# Vorherige (laufende) Sechsecke ausgeben
        self.bit_complete()
        if self.bitcount % 4 == 0:
            self.nibble_complete()  # ausgeben
        if self.bitcount == 16:
            self.ad_complete([3,['%04X' % self.a]])
        if self.bitcount == 28:
            self.ad_complete([4,['%02X' % self.d]])
# Neue Sechsecke beginnen
        if self.bitcount != 19:
            if self.bitcount == 16:
                self.bit = [1,['Even','E'] if dat else ['Odd','O']]
            else:
                self.bit = [0,['%d' % dat]]
            self.ss_bit = self.samplenum    # Startzeit dieses Bits merken (außer Bit 19)
        if self.bitcount % 4 == 0:
            self.nibble = 0
            self.ss_nibble = self.samplenum # Startzeit des Nibbles merken
        self.nibble = (self.nibble<<1) | dat   # Nibble bauen
        if self.bitcount == 0:
            self.a = 0
            self.ss_ad = self.samplenum      # Start eines Adressworts
        if self.bitcount <= 15:
            self.a = self.a<<1 | dat
            if self.bitcount == 15 and self.a&7 != len(self.r)&7: # kontinuierliche Adresse xyz0..xyz7 oder xyz8..xyzF prüfen
                self.ss_r = -1                  # Kein Datensatz / Datensatz abbrechen
        if self.bitcount == 16 and not(bin(self.a).count('1') + dat & 1):
            self.ss_r = -1                  # Kein Datensatz / Datensatz abbrechen
        if self.bitcount == 18 and not dat:
            self.ss_r = -1                  # Kein Datensatz / Datensatz abbrechen
        if self.bitcount == 20:
            self.d = 0
            self.ss_ad = self.samplenum      # Start eines (Antwort?-)Bytes
        if 20 <= self.bitcount <= 27:
            self.d = self.d<<1 | dat
            if self.bitcount == 27 and self.ss_r >= 0:
                self.r.append(self.d)
        self.bitcount += 1
    
    def clk_fall(self, dat):
        if self.bitcount == 20:             # Nur bei Bit 20 erfolgt die Präsenzkennung der Karte auf der Rückflanke
            self.bit = [1,['Nak','N'] if dat else ['Ack','A']]
            self.ss_bit = self.samplenum    # Startzeit dieses Bits merken

    def handle_edges(self, dat):
        if self.inverted:
            dat = 1-dat
        if self.matched[0]:     # Startflanke von SEL
            self.bitcount = 0
            if self.ss_r < 0:
                self.ss_r = self.samplenum
                self.r = []
        if self.matched[1]:     # Endflanke von SEL
            self.bit_complete() # Sechsecke ausgeben
            self.nibble_complete()
            self.ad_complete([4,['%02X' % self.d]])
            self.r_complete()
        if self.matched[2]:     # Steigende Taktflanke
            self.clk_rise(dat)
        if self.matched[3]:     # Fallende Taktflanke
            self.clk_fall(dat)

    def decode(self):
        # {Anfang und Ende von SEL} sowie {Taktflanken solange SEL aktiv} suchen
        high = 'l' if self.inverted else 'h'
        rise = 'f' if self.inverted else 'r'
        fall = 'r' if self.inverted else 'f'

        while True:
            (sel,clk,dat) = self.wait([{0:rise},{0:fall},{0:high,1:rise},{0:high,1:fall}])
            self.handle_edges(dat)
